
class StageType(type):

    def __new__(cls, name, bases, attrs):
        
        # Check for meta-data in the Stage.

        meta = {}

        if 'Meta' in attrs:
            # Extract all data from the clas and accept just the keys
            # that don't start with _.
            data = attrs.pop('Meta').__dict__.items()
            meta.update(dict((k,v) for k,v in data if not k.startswith('_')))
    
        # Add meta to the stage itself.
        attrs['_meta'] = attrs.get('_meta', {})
        attrs['_meta'].update(meta)

        # Decorate the run method.
        def run_decorator(method):
            def new_run(self, *args, **kwargs):
                
                # Ensure run all the requires.
                try:
                    requires = list(self._meta.get('requires', []))
                except TypeError:
                    requires = [self._meta.get('requires')]

                for require in requires:
                    require.run(*args, **kwargs)

                return method(self, *args, **kwargs)

            return new_run
            
        attrs['run'] = run_decorator(attrs.pop('run'))

        return super(StageType, cls).__new__(cls, name, bases, attrs)


class Stage(object):
    """
    Stage
    =====

    A stage will be the way to group different components in a description.

    It will help later to group components and to force a description.

    """

    __metaclass__ = StageType

    def run(self):
        raise NotImplementedError("This stage hasn't been implemented yet.")
    



from configparser import ConfigParser


class Configuration(object):
    """
    Configurations
    ==============

    Configurations are going to be managed in namespaces, that is, we can 
    create a particular namespace in espresso and then look this setting
    later, but any module can use its own namespace.

    The configurations are going to be stored in the filesystem in the path:

       /etc/espresso/conf.d/<namespace>.conf

    Using the configparser module, then using ini-like configs.

    An option is specified then by namespace.section.name and is stored as
    a single string.
    
    """

    def __init__(self, namespace):
        self.namespace = namespace

    def get(self, section, name):
        " Get an option."

        conf = ConfigParser()
        conf.read('/etc/espresso/conf.d/%s.conf' % (self.namespace, ))
        return conf.get(section, name)

    def set(self, section, name, value):
        " Set an option."

        conf = ConfigParser()
        conf.read('/etc/espresso/conf.d/%s.conf' % (self.namespace, ))

        # Check for section
        if not conf.has_section(section):
            conf.add_section(section)
        
        conf.set(section, name, value)

        # Write back configurations.
        with open('/etc/espresso/conf.d/%s.conf' % (self.namespace, ), 'w') as fd:
            conf.write(fd)



import os
import sys
import sh
import pwd
import fstab as fstab_lib

class System(object):

    def ensure_user(self, name, home=None):
        " Ensure the presence of a user in the system and its home path."

        print(" * Check if the user exists ... ")
        
        if any(name == u.pw_name for u in pwd.getpwall()):
            print("   ... user already exists. ")
            return

        print("   ... create user ...")
        
        if home is not None:            
            sh.useradd('-d', home, '-m', name, _err=sys.stderr, _out=sys.stdout)
        else:
            sh.useradd('-m', name, _err=sys.stderr, _out=sys.stdout)
            
        print("   ... done!")


class FsTab(object):

    def remount(self, directory):
        " Remount a directory."

        sh.mount('-o', 'remount', directory, _err=sys.stderr, _out=sys.stdout)

    def ensure_option(self, option, directory):
        " Ensure an option to the mounting at a directory."

        print(" * Ensure that mount line with directory '%s' has option '%s' ..." % (
            directory, option))
        
        fs = fstab_lib.Fstab()
        fs.read('/etc/fstab')

        # Get line with file.
        line = [i for i, line in enumerate(fs.lines) if line.directory == directory]
        if len(line) == 0:
            raise ValueError("Directory is not used by any mount.")

        line = line[0]            

        # Get options.
        options = fs.lines[line].get_options()

        # Option on it?
        if option not in options:

            options.append(option)

            # Set new option.
            fs.lines[line].set_options(options)            
            
            # Write change.
            fs.write('/etc/fstab')

            print("   ... option added ...")
        
        print("   ... done!")
        

class Shell(object):
    def run(self, line):
        sh.bash('-c', line, _err=sys.stderr, _out=sys.stdout)

class DebConf(object):
    def set_selections(self, content):
        print(" * Set selections for debian packages.")
        sh.Command('debconf-set-selections')(_in=content, _err=sys.stderr, _out=sys.stdout)

class Dpkg(object):

    def add_key(self, key):
        print(" * Adding public key to repositories ...")
        sh.Command('apt-key')('add', '-', _in=key, _err=sys.stderr, _out=sys.stdout)
    
    def update(self):
        print(" * Updating dpkg repositories ...")
        sh.aptitude('-y', 'update', _err=sys.stderr, _out=sys.stdout)

    def safe_upgrade(self):
        sh.aptitude('-y', 'safe-upgrade', _err=sys.stderr, _out=sys.stdout)

    def ensure_installed(self, packages):
        sh.aptitude('-y', 'install', *packages, _err=sys.stderr, _out=sys.stdout)


class FileSystem(object):
    " Some basic functions. "

    def ensure_file(self, path, content=None, size=None):
        " Ensure a file exists and its content."

        print(" * Ensuring that '%s' is a file ..." % (path, ))
    
        if os.path.exists(path):
            if not os.path.isfile(path):
                raise TypeError("The path is already used by a non-file.")
                
        elif size > 0:
            with open(path, 'w') as fd:                
                print("   ... ensuring size of %d bytes ..." %(size, ))
                fd.truncate(size)

        # Ensure content.
        if content is not None:
            with open(path, 'w') as fd:
                print("   ... ensuring content in the file ...")
                fd.write(content)                                        

        print("   ... done!")

    def ensure_directory(self, path, mode=None, owner=None, group=None):
        " Ensure a directory exists in a path."
        
        print(" * Ensuring that '%s' is a directory ..." % (path, ))

        if os.path.exists(path):
            if not os.path.isdir(path):
                raise TypeError("The path exists and is not a directory!")
        else:
            # Create the path.
            print("  ... creating ...")
            os.makedirs(path)
                        
        # TODO: Ensure mode, owner and group.
        print("   ... done!")


# Create helpers.
fs = FileSystem()
dpkg =Dpkg()
debconf = DebConf()
shell = Shell()
fstab = FsTab()
system = System()
